#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <signal.h>

#include <vector>
#include <iostream>
#include <iomanip>
#include <stdexcept>

#include "curtain.h"

const int MOTION =    0xc0;
const int TIME_OUT  = 0x02;
const int DOME_OPEN = 0x40;
const int RSWI_OPEN = 0x20;
const int LSWI_OPEN = 0x10;

const int DOME_CLOSE = 0x04;
const int RSWI_CLOSE = 0x02;
const int LSWI_CLOSE = 0x01;

const int TIME_UNITS = 200;

const std::string   sect_device("Devices");

const SComm Curtain::OPEN = 0x80;
const SComm Curtain::CLOSE =0x81;
const SComm Curtain::STOP  =0x85;
const SComm Curtain::UNLOCK = 0x8c;
const SComm Curtain::LOCK = 0x8e;

const SWord Curtain::SET_LIM_TIME =0x32;
const SWord Curtain::SET_LIM_INDEX = 0x34;
const SWord Curtain::SET_ACTIV_PAUSE = 0x36;
const AData Curtain::GET_POSITION = 0xA0;

pthread_t Curtain::Watch_thread;
//      ,   
//    -, /  ..
Curtain::Curtain( Config& cfg, RS485& r, const std::string& s ) : Module( cfg, r, s ) {
    sleep_time = cfg.inp<int>( sect_device, s, "ActivityCheck");
    lim_motion_time = cfg.inp<int>( sect_device, s, "LimitTime");
    freeze = cfg.inp<double>( sect_device, s, "MinSpeed");
    norm_rate = cfg.inp<double>( sect_device, s, "NormalRate");
                                    //   
    sendWordCommand( SET_LIM_TIME, lim_motion_time*TIME_UNITS );
    lineWatchdog();
//    std::cerr << "Constructor Curtain" << std::endl;
}

Curtain::~Curtain() {
//    std::cerr << "Destructor Curtain beg" << std::endl;
    if( pthread_cancel(Watch_thread) == ESRCH ) {
        ErrlogSingl::instance().write("Watch thread is terminated already");
        return;
    }
    if( pthread_join(Watch_thread, NULL) != 0 ) {
        ErrlogSingl::instance().write(ERR_ALG, "Curtain Watch pthread_join failed");
    }
//    std::cerr << "Destructor Curtain end" << std::endl;
}

point<int> Curtain::getPosition() {
    std::vector<unsigned char> x = askData(GET_POSITION);
    if( x.size() != 4 ) throw ErrFat( ERR_EXCHANGE, "Wrong position data" );
    return curr_position = ( x.size() != 4 )? point<int>() :
           point<int>( (int)*(short*)&x[0], (int)*(short*)&x[2] );
}

const std::string& Curtain::devStatus() {
    static const std::string dome_stat[] = { "UNDEFINED", "CLOSED", "OPENED" };
    int istatus = 0;
    if( getSensor()== 0 ) istatus = 0;
    if( DOME_CLOSE&sensor || (RSWI_CLOSE|LSWI_CLOSE)&sensor ) istatus = 1;
    if( DOME_OPEN &sensor || ( RSWI_OPEN| LSWI_OPEN)&sensor ) istatus = 2;
    return dome_stat[istatus];
}

Curtain *Curtain::Handler::curts;

bool Curtain::Handler::expired = false;

void Curtain::Handler::catch_stop(int sig) {
    curts->sendSimpleCommand( STOP );
    ErrlogSingl::instance().write( "Process was externally stopped" );
    expired = true;
}

Curtain::Handler::Handler( Curtain* th) {
    expired = false;
    curts = th;
    signal( SIGUSR1, catch_stop );
}

void Curtain::close(){
    Handler h( this );
    if( DOME_CLOSE&getSensor() ) return;
    if( (RSWI_CLOSE|LSWI_CLOSE)&sensor ) {
        ErrlogSingl::instance().write( "Dome isn't quite closed" );
        return;
    }
    getPosition();
    lim_motion_time = (int)rint(( avgPosition() - 0 )/norm_rate + 5);
    sendWordCommand( SET_LIM_TIME, lim_motion_time*TIME_UNITS );
    sendSimpleCommand( UNLOCK );
    sendSimpleCommand( CLOSE );
    std::queue< point<int> > pos_que;
    while( MOTION&getStatus() && !h.expired ) {
        usleep( 1000000/qlength );
        pos_que.push( getPosition() );
        if( pos_que.size() < qlength ) continue;
        point<int> deltp = pos_que.front() - curr_position;
        pos_que.pop();
        if( deltp.x() < freeze &&  deltp.y() < freeze ) {
            sendSimpleCommand( STOP );
            ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Curtain doesn't move" );
        }
    }
    sendSimpleCommand( LOCK );
    getPosition();
    if( !(DOME_CLOSE&getSensor()) )
        ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Dome isn't quite closed" );
    else  ErrlogSingl::instance().write( "Dome is closed now" );
}

void Curtain::open( int pos_opd){
    Handler h( this );
    sendWordCommand( SET_LIM_INDEX, pos_opd );
    if( DOME_OPEN&getSensor() ) return;
    getPosition();
    lim_motion_time = (int)rint(( pos_opd - avgPosition() )/norm_rate + 5);
    sendWordCommand( SET_LIM_TIME, lim_motion_time*TIME_UNITS );
    sendSimpleCommand( UNLOCK );
    sendSimpleCommand( OPEN );
    std::queue< point<int> > pos_que;
    while( MOTION&getStatus() && !h.expired ) {
        usleep( 1000000/qlength );
        pos_que.push( getPosition() );
        if( pos_que.size() < qlength ) continue;
        point<int> deltp = curr_position - pos_que.front();
        pos_que.pop();
        if( deltp.x() < freeze &&  deltp.y() < freeze ) {
            sendSimpleCommand( STOP );
            ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Curtain doesn't move" );
        }
    }
    sendSimpleCommand( LOCK );
    getPosition();
    if( !(DOME_OPEN&getSensor()) )
        ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Dome isn't quite opened" );
    else ErrlogSingl::instance().write( "Dome is opened now" );
}

void Curtain::lineWatchdog(){
    sendWordCommand( SET_ACTIV_PAUSE, sleep_time*TIME_UNITS );
    pthread_create( &Watch_thread, NULL, &Curtain::thread, (void*)this );
}

void *Curtain::thread( void* a ) {
    Curtain* ptr = (Curtain*)a;
    while(true) {
        try {
            if( ptr->connection() ) ptr->getStatus();       //    sleep_time/2
            nsleep(ptr->sleep_time/2);
        }
        catch ( RS485::ErrFatal& e ) {
            ErrlogSingl::instance().write( ERR_EXCHANGE, e.what() );
//            return 0;           // -  
        }
        catch (const std::exception& e ) {
            ErrlogSingl::instance().write( ERR_STD_EXCP, e.what() );
        }
    }
    return 0;
}
